use std::fmt::Debug;

use actix::Addr;
use actix_identity::Identity;
use actix_web::{HttpRequest, HttpResponse, Responder, web};
use actix_web::error::ErrorUnauthorized;
use askama::Template;

use crate::actors::{openid_sessions_actor, users_actor};
use crate::actors::openid_sessions_actor::{OpenIDSessionsActor, Session, SessionID};
use crate::actors::users_actor::UsersActor;
use crate::constants::*;
use crate::controllers::base_controller::FatalErrorPage;
use crate::data::app_config::AppConfig;
use crate::data::client::{ClientID, ClientManager};
use crate::data::code_challenge::CodeChallenge;
use crate::data::current_user::CurrentUser;
use crate::data::id_token::IdToken;
use crate::data::jwt_signer::{JsonWebKey, JWTSigner};
use crate::data::open_id_user_info::OpenIDUserInfo;
use crate::data::openid_config::OpenIDConfig;
use crate::data::session_identity::SessionIdentity;
use crate::data::user::User;
use crate::utils::string_utils::rand_str;
use crate::utils::time::time;

pub async fn get_configuration(req: HttpRequest, app_conf: web::Data<AppConfig>) -> impl Responder {
    let is_secure_request = req.headers().get("HTTP_X_FORWARDED_PROTO")
        .map(|v| v.to_str().unwrap_or_default().to_lowercase().eq("https"))
        .unwrap_or(false);

    let host = match req.headers().get("Host") {
        None => return HttpResponse::BadRequest().body("Missing Host header!"),
        Some(s) => s.to_str().unwrap_or_default(),
    };

    let curr_origin = format!("{}://{}", match is_secure_request {
        true => "https",
        false => "http"
    }, host);

    HttpResponse::Ok().json(OpenIDConfig {
        issuer: app_conf.website_origin.clone(),
        authorization_endpoint: app_conf.full_url(AUTHORIZE_URI),
        token_endpoint: curr_origin.clone() + TOKEN_URI,
        userinfo_endpoint: curr_origin.clone() + USERINFO_URI,
        jwks_uri: curr_origin + CERT_URI,
        scopes_supported: vec!["openid", "profile", "email"],
        response_types_supported: vec!["code", "id_token", "token id_token"],
        subject_types_supported: vec!["public"],
        id_token_signing_alg_values_supported: vec!["RS256"],
        token_endpoint_auth_methods_supported: vec!["client_secret_post", "client_secret_basic"],
        claims_supported: vec!["sub", "name", "given_name", "family_name", "email"],
        code_challenge_methods_supported: vec!["plain", "S256"],
    })
}

#[derive(serde::Deserialize, Debug)]
pub struct AuthorizeQuery {
    /// REQUIRED. OpenID Connect requests MUST contain the openid scope value. If the openid scope value is not present, the behavior is entirely unspecified. Other scope values MAY be present. Scope values used that are not understood by an implementation SHOULD be ignored. See Sections 5.4 and 11 for additional scope values defined by this specification.
    scope: String,

    /// REQUIRED. OAuth 2.0 Response Type value that determines the authorization processing flow to be used, including what parameters are returned from the endpoints used. When using the Authorization Code Flow, this value is code.
    response_type: String,

    /// REQUIRED. OAuth 2.0 Client Identifier valid at the Authorization Server.
    client_id: ClientID,

    /// REQUIRED. Redirection URI to which the response will be sent. This URI MUST exactly match one of the Redirection URI values for the Client pre-registered at the OpenID Provider, with the matching performed as described in Section 6.2.1 of RFC3986 (Simple String Comparison). When using this flow, the Redirection URI SHOULD use the https scheme; however, it MAY use the http scheme, provided that the Client Type is confidential, as defined in Section 2.1 of OAuth 2.0, and provided the OP allows the use of http Redirection URIs in this case. The Redirection URI MAY use an alternate scheme, such as one that is intended to identify a callback into a native application.
    redirect_uri: String,

    /// RECOMMENDED. Opaque value used to maintain state between the request and the callback. Typically, Cross-Site Request Forgery (CSRF, XSRF) mitigation is done by cryptographically binding the value of this parameter with a browser cookie.
    state: String,

    /// OPTIONAL. String value used to associate a Client session with an ID Token, and to mitigate replay attacks. The value is passed through unmodified from the Authentication Request to the ID Token. Sufficient entropy MUST be present in the nonce values used to prevent attackers from guessing values.
    nonce: Option<String>,

    /// OPTIONAL - <https://ldapwiki.com/wiki/Code_challenge_method>
    code_challenge: Option<String>,
    code_challenge_method: Option<String>,
}

fn error_redirect(query: &AuthorizeQuery, error: &str, description: &str) -> HttpResponse {
    log::warn!("Failed to process sign in request ({} => {}): {:?}", error, description, query);
    HttpResponse::Found()
        .append_header(
            ("Location", format!(
                "{}?error={}?error_description={}&state={}",
                query.redirect_uri,
                urlencoding::encode(error),
                urlencoding::encode(description),
                urlencoding::encode(&query.state)
            ))
        )
        .finish()
}

pub async fn authorize(user: CurrentUser, id: Identity, query: web::Query<AuthorizeQuery>,
                       clients: web::Data<ClientManager>,
                       sessions: web::Data<Addr<OpenIDSessionsActor>>) -> impl Responder {
    let client = match clients.find_by_id(&query.client_id) {
        None => {
            return HttpResponse::BadRequest().body(FatalErrorPage {
                message: "Client is invalid!"
            }.render().unwrap());
        }
        Some(c) => c
    };

    let redirect_uri = query.redirect_uri.trim().to_string();
    if !redirect_uri.starts_with(&client.redirect_uri) {
        return HttpResponse::BadRequest().body(FatalErrorPage {
            message: "Redirect URI is invalid!"
        }.render().unwrap());
    }

    if !query.scope.split(' ').any(|x| x == "openid") {
        return error_redirect(&query, "invalid_request", "openid scope missing!");
    }

    if !query.response_type.eq("code") {
        return error_redirect(&query, "invalid_request", "Only code response type is supported!");
    }

    if query.state.is_empty() {
        return error_redirect(&query, "invalid_request", "State is empty!");
    }

    let code_challenge = match query.0.code_challenge.clone() {
        Some(chal) => {
            let meth = query.0.code_challenge_method.as_deref().unwrap_or("plain");
            if !meth.eq("S256") && !meth.eq("plain") {
                return error_redirect(&query, "invalid_request",
                                      "Only S256 and plain code challenge methods are supported!");
            }
            Some(CodeChallenge { code_challenge: chal, code_challenge_method: meth.to_string() })
        }
        _ => None
    };

    // Check if user is authorized to access the application
    if !user.can_access_app(&client.id) {
        return error_redirect(&query, "invalid_request",
                              "User is not authorized to access this application!");
    }

    // Save all authentication information in memory
    let session = Session {
        session_id: SessionID(rand_str(OPEN_ID_SESSION_LEN)),
        client: client.id,
        user: user.uid.clone(),
        auth_time: SessionIdentity(&id).auth_time(),
        redirect_uri,
        authorization_code: rand_str(OPEN_ID_AUTHORIZATION_CODE_LEN),
        authorization_code_expire_at: time() + OPEN_ID_AUTHORIZATION_CODE_TIMEOUT,
        access_token: None,
        access_token_expire_at: time() + OPEN_ID_ACCESS_TOKEN_TIMEOUT,
        refresh_token: "".to_string(),
        refresh_token_expire_at: 0,
        nonce: query.0.nonce,
        code_challenge,
    };
    sessions.send(openid_sessions_actor::PushNewSession(session.clone())).await.unwrap();

    log::trace!("New OpenID session: {:#?}", session);

    HttpResponse::Found()
        .append_header(("Location", format!(
            "{}?state={}&session_state={}&code={}",
            session.redirect_uri,
            urlencoding::encode(&query.0.state),
            urlencoding::encode(&session.session_id.0),
            urlencoding::encode(&session.authorization_code)
        ))).finish()
}

#[derive(serde::Serialize)]
struct ErrorResponse {
    error: String,
    error_description: String,
}

pub fn error_response<D: Debug>(query: &D, error: &str, description: &str) -> HttpResponse {
    log::warn!("request failed: {} - {} => '{:#?}'", error, description, query);
    HttpResponse::BadRequest()
        .json(ErrorResponse {
            error: error.to_string(),
            error_description: description.to_string(),
        })
}

#[derive(Debug, serde::Deserialize)]
pub struct TokenAuthorizationCodeQuery {
    redirect_uri: String,
    code: String,
    code_verifier: Option<String>,
}

#[derive(Debug, serde::Deserialize)]
pub struct TokenRefreshTokenQuery {
    refresh_token: String,
}


#[derive(Debug, serde::Deserialize)]
pub struct TokenQuery {
    grant_type: String,
    client_id: Option<ClientID>,
    client_secret: Option<String>,

    #[serde(flatten)]
    authorization_code_query: Option<TokenAuthorizationCodeQuery>,

    #[serde(flatten)]
    refresh_token_query: Option<TokenRefreshTokenQuery>,
}

#[derive(Debug, serde::Serialize)]
pub struct TokenResponse {
    access_token: String,
    token_type: &'static str,
    refresh_token: String,
    expires_in: u64,
    #[serde(skip_serializing_if = "Option::is_none")]
    id_token: Option<String>,
}

pub async fn token(req: HttpRequest,
                   query: web::Form<TokenQuery>,
                   clients: web::Data<ClientManager>,
                   app_config: web::Data<AppConfig>,
                   sessions: web::Data<Addr<OpenIDSessionsActor>>,
                   users: web::Data<Addr<UsersActor>>,
                   jwt_signer: web::Data<JWTSigner>) -> actix_web::Result<HttpResponse> {

    // Extraction authentication information
    let authorization_header = req.headers().get("authorization");
    let (client_id, client_secret) = match (&query.client_id, &query.client_secret, authorization_header) {
        // post authentication
        (Some(client_id), Some(client_secret), None) => {
            (client_id.clone(), client_secret.to_string())
        }

        // Basic authentication
        (None, None, Some(v)) => {
            let token = match v.to_str().unwrap_or_default().strip_prefix("Basic ") {
                None => {
                    return Ok(error_response(
                        &query,
                        "invalid_request",
                        &format!("Authorization header does not start with 'Basic ', got '{:#?}'", v),
                    ));
                }
                Some(v) => v
            };

            let decode = String::from_utf8_lossy(&match base64::decode(token) {
                Ok(d) => d,
                Err(e) => {
                    log::error!("Failed to decode authorization header: {:?}", e);
                    return Ok(error_response(&query, "invalid_request", "Failed to decode authorization header!"));
                }
            }).to_string();

            match decode.split_once(':') {
                None => (ClientID(decode), "".to_string()),
                Some((id, secret)) => (ClientID(id.to_string()), secret.to_string())
            }
        }

        _ => {
            return Ok(error_response(&query, "invalid_request", "Authentication method unknown!"));
        }
    };

    let client = clients
        .find_by_id(&client_id)
        .ok_or_else(|| ErrorUnauthorized("Client not found"))?;

    if !client.secret.eq(&client_secret) {
        return Ok(error_response(&query, "invalid_request", "Client secret is invalid!"));
    }

    let token_response = match (query.grant_type.as_str(),
                                &query.authorization_code_query,
                                &query.refresh_token_query) {
        ("authorization_code", Some(q), _) => {
            let mut session: Session = match sessions
                .send(openid_sessions_actor::FindSessionByAuthorizationCode(q.code.clone()))
                .await.unwrap()
            {
                None => {
                    return Ok(error_response(&query, "invalid_request", "Session not found!"));
                }
                Some(s) => s,
            };

            if session.client != client.id {
                return Ok(error_response(&query, "invalid_request", "Client mismatch!"));
            }

            if session.redirect_uri != q.redirect_uri {
                return Ok(error_response(&query, "invalid_request", "Invalid redirect URI!"));
            }

            if session.authorization_code_expire_at < time() {
                return Ok(error_response(&query, "invalid_request", "Authorization code expired!"));
            }

            // Check code challenge, if needed
            if let Some(chall) = &session.code_challenge {
                let code_verifier = match &q.code_verifier {
                    None => {
                        return Ok(error_response(&query, "access_denied", "Code verifier missing"));
                    }
                    Some(s) => s
                };

                if !chall.verify_code(code_verifier) {
                    return Ok(error_response(&query, "invalid_grant", "Invalid code verifier"));
                }
            } else if q.code_verifier.is_some() {
                return Ok(error_response(&query, "invalid_grant", "Unexpected `code_verifier` parameter!"));
            }

            if session.access_token.is_some() {
                return Ok(error_response(&query, "invalid_request", "Authorization code already used!"));
            }

            session.regenerate_access_and_refresh_tokens(&app_config, &jwt_signer)?;

            sessions.send(openid_sessions_actor::UpdateSession(session.clone()))
                .await.unwrap();

            let user: Option<User> = users.send(users_actor::GetUserRequest(session.user.clone()))
                .await.unwrap().0;
            let user = match user {
                None => return Ok(error_response(&query, "invalid_request", "User not found!")),
                Some(u) => u,
            };

            // Generate id token
            let id_token = IdToken {
                issuer: app_config.website_origin.to_string(),
                subject_identifier: session.user.0,
                audience: session.client.0.to_string(),
                expiration_time: session.access_token_expire_at,
                issued_at: time(),
                auth_time: session.auth_time,
                nonce: session.nonce,
                email: user.email,
            };

            TokenResponse {
                access_token: session.access_token.expect("Missing access token!"),
                token_type: "Bearer",
                refresh_token: session.refresh_token,
                expires_in: session.access_token_expire_at - time(),
                id_token: Some(jwt_signer.sign_token(id_token.to_jwt_claims())?),
            }
        }

        ("refresh_token", _, Some(q)) => {
            let mut session: Session = match sessions
                .send(openid_sessions_actor::FindSessionByRefreshToken(q.refresh_token.clone()))
                .await.unwrap()
            {
                None => {
                    return Ok(error_response(&query, "invalid_request", "Session not found!"));
                }
                Some(s) => s,
            };

            if session.client != client.id {
                return Ok(error_response(&query, "invalid_request", "Client mismatch!"));
            }

            if session.refresh_token_expire_at < time() {
                return Ok(error_response(&query, "access_denied", "Refresh token has expired!"));
            }

            session.regenerate_access_and_refresh_tokens(&app_config, &jwt_signer)?;

            sessions
                .send(openid_sessions_actor::UpdateSession(session.clone()))
                .await.unwrap();

            TokenResponse {
                access_token: session.access_token.expect("Missing access token!"),
                token_type: "Bearer",
                refresh_token: session.refresh_token,
                expires_in: session.access_token_expire_at - time(),
                id_token: None,
            }
        }

        _ => {
            return Ok(error_response(&query, "invalid_request", "Grant type unsupported!"));
        }
    };

    Ok(HttpResponse::Ok()
        .append_header(("Cache-Control", "no-store"))
        .append_header(("Pragam", "no-cache"))
        .json(token_response))
}

#[derive(serde::Serialize)]
struct CertsResponse {
    keys: Vec<JsonWebKey>,
}

pub async fn cert_uri(jwt_signer: web::Data<JWTSigner>) -> impl Responder {
    HttpResponse::Ok().json(CertsResponse { keys: vec![jwt_signer.get_json_web_key()] })
}

fn user_info_error(err: &str, description: &str) -> HttpResponse {
    HttpResponse::Unauthorized()
        .insert_header(("WWW-Authenticate", format!(
            "Bearer error=\"{}\", error_description=\"{}\"",
            err,
            description
        )))
        .finish()
}

#[derive(serde::Deserialize)]
pub struct UserInfoQuery {
    access_token: Option<String>,
}

pub async fn user_info_post(req: HttpRequest,
                            form: Option<web::Form<UserInfoQuery>>,
                            query: web::Query<UserInfoQuery>,
                            sessions: web::Data<Addr<OpenIDSessionsActor>>,
                            users: web::Data<Addr<UsersActor>>) -> impl Responder {
    user_info(req,
              form
                  .map(|f| f.0.access_token)
                  .unwrap_or_default()
                  .or(query.0.access_token),
              sessions,
              users,
    ).await
}

pub async fn user_info_get(req: HttpRequest, query: web::Query<UserInfoQuery>,
                           sessions: web::Data<Addr<OpenIDSessionsActor>>,
                           users: web::Data<Addr<UsersActor>>) -> impl Responder {
    user_info(req, query.0.access_token, sessions, users).await
}

/// Authenticate request using RFC6750 <https://datatracker.ietf.org/doc/html/rfc6750>///
async fn user_info(req: HttpRequest, token: Option<String>,
                   sessions: web::Data<Addr<OpenIDSessionsActor>>,
                   users: web::Data<Addr<UsersActor>>) -> impl Responder {
    let token = match token {
        Some(t) => t,
        None => {
            let token = match req.headers().get("Authorization") {
                None => return user_info_error("invalid_request", "Missing access token!"),
                Some(t) => t
            };

            let token = match token.to_str() {
                Err(_) => return user_info_error("invalid_request", "Failed to decode token!"),
                Ok(t) => t,
            };

            let token = match token.strip_prefix("Bearer ") {
                None => return user_info_error("invalid_request", "Header token does not start with 'Bearer '!"),
                Some(t) => t,
            };

            token.to_string()
        }
    };

    let session: Option<Session> = sessions
        .send(openid_sessions_actor::FindSessionByAccessToken(token)).await.unwrap();
    let session = match session {
        None => {
            return user_info_error("invalid_request", "Session not found!");
        }
        Some(s) => s,
    };

    if session.access_token_expire_at < time() {
        return user_info_error("invalid_request", "Access token has expired!");
    }

    let user: Option<User> = users.send(users_actor::GetUserRequest(session.user)).await.unwrap().0;
    let user = match user {
        None => {
            return user_info_error("invalid_request", "Failed to extract user information!");
        }
        Some(u) => u,
    };

    HttpResponse::Ok()
        .json(OpenIDUserInfo {
            name: user.full_name(),
            sub: user.uid.0,
            given_name: user.first_name,
            family_name: user.last_name,
            preferred_username: user.username,
            email: user.email,
            email_verified: true,
        })
}